import numpy as np
from statistics import mode
from .supervised import DecisionTree,RegresiLinear
from ..datamanipulasi.fold import train_val_split

class Bagging:
    def __init__(self,model, n_model:int = 5,random_state:int=None):
        self.__model = model
        self.__label:np.ndarray = None
        self.__random_state = random_state
        self.__n_model = n_model
    
    def fit(self,x:np.ndarray,y:np.ndarray):

        if self.__random_state :
            np.random.seed(self.__random_state)
        record_x = [i for i in range(len(x))]
        predict_model = []

        for i in range(self.__n_model):
            data_x = []
            data_y = []
            for _  in range(len(x)):
                rd_c = np.random.choice(record_x,1,replace=True)
                data_x.append(x[rd_c])
                data_y.append(int(y[rd_c]))
            
            data_x = np.array(data_x).reshape(x.shape)
            data_y = np.array(data_y)
            obj = self.__model
            obj.fit(data_x,data_y)
            predict_model.append(obj)
            pred = obj.predict(x)
            print(
                f"model-{i+1} akurasi : {self.score_accuracy(pred,y)}"
            )

        
        predict = []
        for data in x:
            temp_predict = []
            for m in predict_model:
                temp_predict.append(m.predict(data))
            
            predict.append(mode(temp_predict))


        
        self.__label = np.array(predict)
    
    def fit_predict(self,x:np.ndarray,y:np.ndarray):
        self.fit(x,y)
        return self.__label
    
    def score_accuracy(self,y_pred:np.ndarray,y_true:np.ndarray):
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true/len(y_pred)


class Boosting:

    def __init__(self,model, n_model:int = 5,random_state:int=None):
        self.__model = model
        self.__n_model = n_model
        self.__random_state = random_state
        self.__label = None
    
    def fit(self,x:np.ndarray,y:np.ndarray):
        sample_weight = np.ones(len(x))/len(x)

        if self.__random_state :
            np.random.seed(self.__random_state)
        predict_model = []
        weight_model = []
        final_prediksi = None

        for i in range(self.__n_model):
            selected_indices = np.random.choice(len(x), size=len(x), p=sample_weight)
            data_x = [x[i] for i in selected_indices]
            data_y = [y[i] for i in selected_indices]
            data_x = np.array(data_x)
            data_y = np.array(data_y)
            obj = self.__model
            obj.fit(data_x,data_y)
            predict_model.append(obj)
            prediksi = obj.predict(x)
            error = np.sum(sample_weight * (prediksi != y))
            model_weight = 0.5 * np.log((1 - error) / error)
            weight_model.append(model_weight)
            sample_weight *= np.exp(-model_weight * y * prediksi)
            sample_weight /= np.sum(sample_weight)
            if final_prediksi == None:
                final_prediksi = model_weight*prediksi
            else:
                final_prediksi += model_weight*prediksi
            
            print(
                f"model-{i+1} akurasi : {self.score_accuracy(prediksi,y)}"
            )
            
        self.__label = np.argmax(final_prediksi,axis=1)
    
    def fit_predict(self,x:np.ndarray,y:np.ndarray):
        self.fit(x,y)
        return self.__label

    def score_accuracy(self,y_pred:np.ndarray,y_true:np.ndarray):
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true/len(y_pred)

class RandomForest:
    def __init__(self,max_feature:int,min_fitur:int=0,max_depth:int=2,random_state:int=None,n_tree:int=3):
        self.__max_depth = max_depth
        self.__random_state = random_state
        self.__n_tree = n_tree
        self.__tree = []
        self.__label = None
        self.__max_feature = max_feature
        self.__min_fitur = min_fitur
    
    @property
    def label(self):
        #ini hanya dekorator
        pass

    @label.getter
    def label__(self):
        return self.__label

    @property
    def tree(self):
        #ini hanya dekorator
        pass
    @tree.getter
    def tree__(self):
        return self.__tree
    
    def fit(self, x:np.ndarray,y:np.ndarray,nama_fitur:list=None):


        if self.__random_state :
            np.random.seed(self.__random_state)
        record_x = [i for i in range(len(x))]

        for i in range(self.__n_tree):
            #boosting data
            data_x = []
            data_y = []
            data_terpilih = []
            for _  in range(len(x)):
                rd_c = np.random.choice(record_x,1,replace=True)
                
                data_x.append(x[rd_c])
                data_y.append(int(y[rd_c]))
            data_x = np.array(data_x).reshape(x.shape)
            data_y = np.array(data_y)
            obj = DecisionTree(self.__max_depth)
            obj.fit(data_x,data_y,self.__max_feature,self.__min_fitur)
            self.__tree.append(obj)
            data_x_oob = np.array([x[i] for i in range(len(x)) if i not in data_terpilih])
            data_y_oob = np.array([y[i] for i in range(len(x)) if i not in data_terpilih])
            
            pred = obj.predict(data_x_oob)
            if nama_fitur != None:
                print(
                f"model-{i+1} akurasi : {self.score_accuracy(pred,data_y_oob)} dan OOB error : {1-self.score_accuracy(pred,data_y_oob)} , kolom root : {nama_fitur[obj.root]}"
                )
            else:
                print(
                    f"model-{i+1} akurasi : {self.score_accuracy(pred,data_y_oob)} dan OOB error : {1-self.score_accuracy(pred,data_y_oob)}"
                )
        

    def fit_predict(self,x:np.ndarray,y:np.ndarray,nama_fitur=None):
        self.fit(x,y,nama_fitur)

        predict = []
        for data in x:
            temp_predict = []
            for m in self.__tree:
                temp_predict.append(m.predict(data))
            
            predict.append(mode(temp_predict))

        self.__label = np.array(predict)
        return self.__label
    
    def predict(self,x:np.ndarray):
        predict = []
        for data in x:
            temp_predict = []
            for m in self.__tree:
                temp_predict.append(m.predict(data))
            
            predict.append(mode(temp_predict))

        self.__label = np.array(predict)
        return self.__label
    

    def score_accuracy(self,y_pred:np.ndarray,y_true:np.ndarray):
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true/len(y_pred)

class Stacking:
    def __init__(self,base_model:list,meta_model,val_size=0.2,random_state=None):
        self.__base_model = base_model
        self.__meta_model = meta_model
        self.__label = None
        self.__val_size = val_size
        self.__random_state=random_state
    
    @property
    def label(self):
        #hanya konstruktor
        pass

    @label.getter
    def label__(self):
        return self.__label
    
    def fit(self,x:np.ndarray,y:np.ndarray):
        X_train, X_val, y_train, y_val = train_val_split(x,y,self.__val_size,self.__random_state)
        pred_base_model = []
        for m in self.__base_model:
            pred = m.fit_predict(X_train,y_train)
            print(
                f"Model {m.name} akurasi : {self.score_accuracy(pred,y_train)} "
            )
            pred_base_model.append(m.predict(X_val))
            
        
        data_train_meta = np.column_stack(pred_base_model)
        prediksi = self.__meta_model.fit_predict(data_train_meta,y_val)
        self.__label = prediksi
    
    def fit_predict(self,x:np.ndarray,y:np.ndarray):
        self.fit(x,y)
        return self.__label
    
    def predict(self,x:np.ndarray):
        pred_base_model = []
        for m in self.__base_model:
            pred = m.predict(x)
            pred_base_model.append(pred)
            
        data_train_meta = np.column_stack(pred_base_model)
        prediksi = self.__meta_model.predict(data_train_meta)
        
        return prediksi
        

    def score_accuracy(self,y_pred:np.ndarray,y_true:np.ndarray):
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true/len(y_pred)
                

class BoostingRegressi:
    def __init__(self,learning_rate:float=0.2,n_estimator=10):
        self.__learning_rate = learning_rate
        self.__n_estimator = n_estimator
        self.__model = []
        self.__mean_target = None
        self.__y = None
    
    @property
    def model(self):
        #ini hanya dekorator
        pass

    @model.getter
    def model_(self):
        return self.__model

    
    def fit(self,x:np.ndarray,y:np.ndarray):
        self.__mean_target = np.mean(y)
        self.__y = y
        prediksi = np.full_like(y,np.mean(y))

        for _ in range(self.__n_estimator):
            residual = y - prediksi
            linear_model = RegresiLinear()
            linear_model.fit(x,residual)
            
            prediksi += self.__learning_rate * linear_model.predict(x)
            self.__model.append(linear_model)
    
    def fit_predict(self,x:np.ndarray,y:np.ndarray):
        self.fit(x,y)
        pred = np.full_like(y,self.__mean_target)
        for mod in self.__model:
            pred += (self.__learning_rate * mod.predict(x))
        return pred

    def predict(self,x:np.ndarray):
        pred = np.full_like(x,shape=(x.shape[0],),fill_value=self.__mean_target)
        for mod in self.__model:
            pred += (self.__learning_rate * mod.predict(x))
        if len(x.shape)==1:
            pred = pred[0]
        return pred
    



        











            
            
            


            





        



import numpy as np

def train_val_split(x:np.ndarray, y:np.ndarray, val_size=0.2, random_state=None):
    if random_state is not None:
        np.random.seed(random_state)

    data_size = len(x)
    val_size = int(val_size * data_size)

    indices = np.random.permutation(data_size)
    train_indices, val_indices = indices[val_size:], indices[:val_size]

    X_train, X_val = x[train_indices], x[val_indices]
    y_train, y_val = y[train_indices], y[val_indices]

    return X_train, X_val, y_train, y_val


"""
Kumpulan algoritma unsupervised learning
"""
import numpy as np
import seaborn as sns
from ..matematika.distance import euclidian_distance,kuadrat_jarak

class Kmeans:

    def __init__(self,cluster:int=2,inital_centroid=None,limit_loop:int=300,konvergen=None,random_state:int=2):
        self.__cluster = cluster
        self.__inital_centroid=inital_centroid
        self.__limit_loop=limit_loop
        self.__historycentroid = {}
        self.__konvergen = konvergen
        self.__random_state = random_state
        self.__cluster_fit = None
        self.__inersia = None
        self.__x = None
        self.__label = None
    
    @property
    def inersia(self):
        #ini hanya sebuah dekorator
        pass

    @property
    def label(self):
        #inin hanya sbeuah dekorator
        pass
    
    def sethistory_centroid(self,loop,list_centroid):
        self.__historycentroid[loop]=list_centroid
    
    def history_centroid(self)->dict:
        return self.__historycentroid
    
    def fit_predict(self,x:np.ndarray)->np.ndarray:
        self.__x = x
        self.fit(x)
        predict = [None for _ in range(len(x))]
        for key3 in self.__cluster_fit.keys():
            for index in self.__cluster_fit[key3]:
                predict[index] = key3
        self.__label = np.array(predict)
        return np.array(predict)

    def fit(self,x:np.ndarray):
        self.__x = x
        np.random.seed(self.__random_state)
        if self.__inital_centroid== None:
            self.__inital_centroid= {}
            for i in range(self.__cluster):
                #centroid = np.random.uniform(x.min(),high = x.max(),size=len(x[0]) )
                centroid = x[np.random.choice(x.shape[0], 1, replace=False)]

                self.__inital_centroid[i] = list(centroid)        
        loop = 0 
        konv = 0
        while loop < self.__limit_loop:
            cluster={}
            for i in range(self.__cluster):
                cluster[i]=[]
            self.sethistory_centroid(loop,self.__inital_centroid)
            for index,value in enumerate(x):
                cls = None
                jrk = None
                for key in self.__inital_centroid.keys():
                    jarak = euclidian_distance(value,np.array(self.__inital_centroid[key]))
                    if jrk == None or jarak <= jrk:
                        jrk = jarak
                        cls = key
                cluster[cls].append(index)
            update_centroid = {}

            for key2 in cluster.keys():
                if len(cluster[key2])!=0:
                    init = x[cluster[key2][0]].copy()
                    for rec in cluster[key2][1:]:
                        init += x[rec].copy()
                    init = init/len(cluster[key2])
                else:
                    init = list( x[np.random.choice(x.shape[0], 1, replace=False)])

                update_centroid[key2] = list(init)
            
            if konv == self.__konvergen:
                break

            if update_centroid == self.__inital_centroid:
                konv +=1
            else:
                self.__inital_centroid = update_centroid
            loop += 1
        
        self.__cluster_fit = cluster

    @inersia.getter
    def inersia_(self):
        total = 0
        
        for key in self.__cluster_fit.keys():
            centroid = self.__inital_centroid[key]
            for rec in self.__cluster_fit[key]:
                total += kuadrat_jarak(self.__x[rec],np.array(centroid))
        self.__inersia = total
        return self.__inersia

    @label.getter
    def label__(self):
        return self.__label 


class DBscan:
    def __init__(self,eps:float, minpts:int):
        self.__eps = eps
        self.__minpts = minpts
        self.__cluster = {}
        self.__label = None
        self.__jarak = []
    
    @property
    def label(self):
        #ini hanya dekorator
        pass
    
    @property
    def jarak(self):
        #ini hanya dekorator
        pass
    
    def fit(self,x:np.ndarray):
        cluster = 1
        visited = []
        unvisited = [i for i in range(len(x))]
        noise = []
        N = []
        while len(unvisited) != 0:
            init = np.random.choice(unvisited, 1, replace=False)[0]
            visited.append(init)
            unvisited.remove(init)
            elemen = []
            for index,value in enumerate(x):
                if index != init :
                    jarak = euclidian_distance(x[init],value)
                    self.__jarak.append(jarak)
                    if jarak <= self.__eps:
                        N.append(index)
            
            if len(N) >= self.__minpts:
                elemen.append(init)
            else:
                N = []
                noise.append(init)
            
            while len(N) >0:
                
                if N[0] not in visited:
                    visited.append(N[0])
                    unvisited.remove(N[0])
                el2 = []
                for idx2,value in enumerate(x):
                    if N[0] != idx2:
                        jarak = euclidian_distance(x[N[0]],value)
                        self.__jarak.append(jarak)
                        if jarak <= self.__eps:
                            el2.append(idx2)
                if len(el2)>= self.__minpts:
                    for j in el2:
                        if j not in visited and j not in N and j not in noise:
                            N.append(j)
                            
                found = False
                for lst in self.__cluster.values():
                    
                    if N[0] in lst:
                        found = True
                        break
                if not found  :
                    elemen.append(N[0])

                N.pop(0)
            
            if len(elemen) > 0:
                self.__cluster[cluster] = elemen
                cluster += 1
        self.__cluster[-1] = noise
    
    def fit_predict(self,x:np.ndarray):
        self.fit(x.copy())
        predict = [None for _ in range(len(x))]
        for key in self.__cluster.keys():
            for index in self.__cluster[key]:
                predict[index] = key
        self.__label = np.array(predict)
        return self.__label
    
    @label.getter
    def label_(self):
        return self.__label
    
    @jarak.getter
    def jarak_(self):
        return self.__jarak

    
                    

                





                

            

            




    

    



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.metrics import silhouette_score
from random import sample
sns.set()

def elbow_method(model,start_cluster:int,end_cluster:int,data:np.ndarray):
    x = []
    y = []

    for x_ in range(start_cluster,end_cluster):
        
        k = model(x_)
        k.fit(data)
        insersia = k.inersia_
        x.append(x_)
        y.append(insersia)
    
    plt.plot(x,y,marker='o',linestyle='--')
    plt.title("Optimal Jumlah Cluster")
    plt.xlabel("cluster")
    plt.ylabel("Sum Square Error")
    plt.tight_layout()
    plt.show()

def hopkins_statistik(x:np.ndarray,n_neighbors:int=2):
    
    n = len(x)
    nei = NearestNeighbors(n_neighbors=n_neighbors).fit(x)

    sli = sample(range(0,n,1),n)
    rand_data = [x[i] for i in sli]
    u_distances = nei.kneighbors(x, n_neighbors=n_neighbors)[0][:, -1]
    w_distances = nei.kneighbors(rand_data, n_neighbors=n_neighbors)[0][:, -1]

    hop = np.sum(u_distances)/(np.sum(u_distances)+np.sum(w_distances))

    return hop

def visual_silhoutte(model,start_cluster:int,end_cluster:int,data:np.ndarray):
    no_cluster = []
    siluette_score = []
    for n in range(start_cluster,end_cluster):
        k = model(n)
        k.fit_predict(data)
        label = k.label__
        siluette_score.append(silhouette_score(data,label))
        no_cluster.append(n)
    
    plt.plot(no_cluster,siluette_score,marker='o',linestyle='--')
    plt.title("Optimal Jumlah Cluster")
    plt.xlabel("cluster")
    plt.ylabel("Silhouette Score")
    plt.tight_layout()
    plt.show()

def visual_silhoutte_dbms(model,title:str,xlabel:str,start:int,end:int,step:int,data:np.ndarray,minpts:int=None,epsilon:int=None):
    no_cluster = []
    siluette_score = []
    if epsilon and not minpts:
        for n in range(start,end,step):
            k = model(epsilon,n)
            k.fit_predict(data)
            label = k.label_
            siluette_score.append(silhouette_score(data,label))
            no_cluster.append(n)
    elif minpts and not epsilon:
        while start!=end:
            k = model(start,minpts)
            k.fit_predict(data)
            label = k.label_
            siluette_score.append(silhouette_score(data,label))
            no_cluster.append(start)
            start += step
    
    plt.plot(no_cluster,siluette_score,marker='o',linestyle='--')
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel("Silhouette Score")
    plt.tight_layout()
    plt.show()

def silhoutte_skor(label:np.ndarray,data:np.ndarray):
    return silhouette_score


def mean_absolute_error(y_true:np.ndarray,y_pred:np.ndarray):
    delta = y_pred - y_true
    absolut = np.abs(delta)
    total = np.sum(absolut)
    return total/len(y_true)

def mean_square_error(y_true:np.ndarray,y_pred:np.ndarray):
    delta = (y_pred - y_true)**2
    total = np.sum(delta)
    return total/len(y_true)

def root_mean_square_eroor(y_true:np.ndarray,y_pred:np.ndarray):
    mse = mean_square_error(y_true,y_pred)
    return np.sqrt(mse)

def mean_percentage_error(y_true:np.ndarray,y_pred:np.ndarray):
    delta = (y_pred - y_true)/y_pred
    absolut = np.abs(delta)
    total = np.sum(absolut)
    return (total/len(y_pred))*100


import numpy as np

class TreeNode:

    def __init__(self, feature=None, value=None, result=None, left=None, right=None):
        self.feature = feature  # Fitur pemisah
        self.value = value      # Nilai pemisah
        self.result = result    # Hasil jika node ini adalah leaf node
        self.left = left        # Sub-tree kiri (nilai kurang dari atau sama dengan pemisah)
        self.right = right      # Sub-tree kanan (nilai lebih dari pemisah)
    
    


import numpy as np

def euclidian_distance(x:np.ndarray,y:np.ndarray)->float:
    delta = x-y
    square = delta**2  
    return np.sqrt(np.sum(square))

def kuadrat_jarak(x:np.ndarray,y:np.ndarray)->float:
    return euclidian_distance(x,y)**2


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.decomposition import PCA
sns.set()

class Visualisasi:
    """
    class untuk visualisasi data
    dengan konstruktor dataframe pandas
    """
    def __init__(self,data):
        self.__data = data
    
    def setData(self,data):
        self.__data = data
    
    def getData(self):
        return self.__data
    
    def distribusi(self):
        num_kolom = len(self.__data.columns)
        num_rows = (num_kolom + 4)//5
        num_rows = min(15, num_rows)

        fig,axes = plt.subplots(num_rows, 5, figsize=(30, 20))
    
        for i, kolom in enumerate(self.__data.columns):
            row = i //5
            col = i % 5
            sns.histplot(self.__data[kolom], kde=True, ax=axes[row, col])
            #axes[row,col].set_title(kolom)
    
        for i in range(num_kolom, num_rows * 5):
            row = i // 5
            col = i % 5
            fig.delaxes(axes[row, col])
    
        fig.suptitle('Distribusi untuk Setiap Kolom', y=1.02)


        plt.tight_layout()
        plt.show()

    def box_plot(self,kolom):
         #fungsi boxplot
        sns.set(rc={'figure.figsize':(38,10)})
        melted_data = pd.melt(self.__data, value_vars=kolom, var_name="variabel", value_name="value")
        ax = sns.boxplot(x="variabel",y="value",data=melted_data)
        ax.set_xticklabels(ax.get_xticklabels(), rotation=90)
        plt.show()
    
    def scatter(self,kolom):
        num_kolom = len(self.__data.columns)
        num_rows = (num_kolom + 4)//5
        num_rows = min(15, num_rows)

        fig,axes = plt.subplots(num_rows, 5, figsize=(30, 20))

        for i, colu in enumerate(self.__data.columns):
            row = i // 5
            col = i % 5
            axes[row, col].scatter(x=self.__data[colu], y=self.__data[kolom])
            axes[row, col].set_xlabel(colu)

        for i in range(num_kolom, num_rows * 5):
            row = i // 5
            col = i % 5
            fig.delaxes(axes[row, col])
    
        fig.suptitle(f'Scatter plot untuk Setiap Kolom terhadap {kolom}', y=1.02)


        plt.tight_layout()
        return plt
    
    def heatmapkor(self):
        korelasi = self.__data.corr()
        plt.figure(figsize=(8, 6))
        sns.heatmap(korelasi, annot=True, cmap='coolwarm', linewidths=.5)

        plt.title('Heatmap Korelasi')
        plt.show()
    
    def regplot(self,colom_x):
        fig, axes = plt.subplots(1, len(self.__data.columns), figsize=(15, 5))
        print(axes)

        for index,y in enumerate(self.__data.columns):
            
            sns.regplot(x=colom_x, y=y, data=self.__data, ci=None, ax=axes[index])
        
        plt.tight_layout
        plt.show()

    

    
class Skalasisasi:
    """
    kelas untuk skalasisasi
    """
    def __init__(self,data):
        self.__data = data.copy()
    
    def minmax_scaler(self):
        kolom = self.__data.columns
        data = self.__data.copy()
        for col in kolom:
            maksimal = data[col].max()
            minimum = data[col].min()
            delta = maksimal - minimum
            data[col] = data[col].apply(lambda x : abs((x-minimum))/delta)
        
        return data
    
    def standar_scaler(self):
        kolom =self.__data.columns
        data = self.__data.copy()
        for col in kolom:
            mean = data[col].mean()
            std = data[col].std()
            data[col] = data[col].apply(lambda x : (x-mean)/std)
        
        return data

def visual_data(tipe:str,data:np.ndarray,Label:np.ndarray,ruang:int=2,legend:list=None):
    data = data.copy()
    pca = PCA(n_components=ruang)
    data = pca.fit_transform(data)
    df = pd.DataFrame(data=data, columns=[f"PCA{i+1}" for i in range(ruang)])
    df['label'] = list(Label)
    print(df)

    scatter = plt.scatter(df['PCA1'], df['PCA2'], c=df['label'], cmap='viridis', s=80)

    plt.title(f'Plot Clustering {tipe} ')
    plt.xlabel('PCA 1')
    plt.ylabel('PCA 2')

    # Menambahkan legenda
    if legend:
        plt.legend(handles=scatter.legend_elements()[0], labels=legend, title='Klaster', loc='upper right')
    else:
        plt.legend(*scatter.legend_elements(), title='Klaster', loc='upper right')
    plt.show()





import numpy as np
from ..graph.tree import TreeNode

class DecisionTree(TreeNode):
    def __init__(self,depth:int=2):
        self.__depth = depth
        self.__tree:TreeNode = None
        self.__label:np.ndarray = None
        self.root = None
    
    @property
    def label(self):
        pass

    @property
    def tree(self):
        pass

    @tree.getter
    def tree__(self):
        return self.__tree
    
    @label.getter
    def label__(self):
        return self.__label
    

    
    def calculate_gini(self,y):
    # Menghitung nilai Gini impurity dari suatu set data
        classes, counts = np.unique(y, return_counts=True)
        probabilities = counts / len(y)
        gini = 1 - np.sum(probabilities**2)
        return gini
    
    def calculate_split_info(self, X:np.ndarray, y:np.ndarray, feature, value):
    # Menghitung nilai split info dari suatu pemisah
        m = len(y)
        X_left, y_left, X_right, y_right = self.split_data(X, y, feature, value)
        p_left = len(y_left) / m
        p_right = len(y_right) / m

        if p_left == 0 or p_right == 0:
            return 0
        split_info = -p_left * np.log2(p_left) - p_right * np.log2(p_right)
        return split_info
    
    def calculate_gain_ratio(self,X:np.ndarray, y:np.ndarray, feature, value):
    # Menghitung nilai Gain Ratio dari suatu pemisah
        gini_before = self.calculate_gini(y)
        m = len(y)

        X_left, y_left, X_right, y_right = self.split_data(X, y, feature, value)

        gini_left = self.calculate_gini(y_left)
        gini_right = self.calculate_gini(y_right)

        gini_after = (len(y_left) / m) * gini_left + (len(y_right) / m) * gini_right

        split_info = self.calculate_split_info(X, y, feature, value)

        if split_info == 0:
            return 0

        gain_ratio = (gini_before - gini_after) / split_info
        return gain_ratio

    def split_data(self,X:np.ndarray, y:np.ndarray, feature, value):
        # Memisahkan data berdasarkan suatu fitur dan nilai pemisah
        mask = X[:, feature] <= value
        X_left, y_left = X[mask], y[mask]
        X_right, y_right = X[~mask], y[~mask]
        return X_left, y_left, X_right, y_right

    def find_best_split(self,X:np.ndarray, y:np.ndarray,min_fitur=0,max_fitur=None):
        # Mencari pemisah terbaik berdasarkan nilai Gain Ratio
        m, n = X.shape
        best_gain_ratio = 0
        best_feature = None
        best_value = None
        if max_fitur == None:
            fitur = [ i for i in range(n)]
        else:
            if min_fitur != 0:
                max_fitur = np.random.randint(min_fitur,max_fitur)
            fitur = np.random.choice(n, size=max_fitur, replace=False)
        

        for feature in fitur:
            unique_values = np.unique(X[:, feature])
            for value in unique_values:
                gain_ratio = self.calculate_gain_ratio(X, y, feature, value)

                if gain_ratio > best_gain_ratio:
                    best_gain_ratio = gain_ratio
                    best_feature = feature
                    best_value = value

        return best_feature, best_value
    
    def build_tree(self,X:np.ndarray, y:np.ndarray, depth=0, max_depth=None,max_feature=None,min_fitur=0):
        # Membangun pohon keputusan secara rekursif
        if depth == max_depth or len(np.unique(y)) == 1:
        # Jika mencapai kedalaman maksimum atau semua label sama, buat leaf node
            return TreeNode(result=np.bincount(y).argmax())

        best_feature, best_value = self.find_best_split(X, y,min_fitur,max_feature)
        if depth == 0:
            self.root = best_feature

        if best_feature is None:
            # Jika tidak dapat menemukan pemisah, buat leaf node
            return TreeNode(result=np.bincount(y).argmax())

        X_left, y_left, X_right, y_right = self.split_data(X, y, best_feature, best_value)

        left_subtree = self.build_tree(X_left, y_left, depth + 1, max_depth,max_feature,min_fitur)
        right_subtree = self.build_tree(X_right, y_right, depth + 1, max_depth,max_feature,min_fitur)

        return TreeNode(feature=best_feature, value=best_value, left=left_subtree, right=right_subtree)
    
    def fit(self,X:np.ndarray,y:np.ndarray,max_feature=None,min_fitur=0):
        self.__tree = self.build_tree(X,y,max_depth=self.__depth,max_feature=max_feature,min_fitur=min_fitur)
        label = []
        for sam in X:
            label.append(self.predict_tree(self.__tree,sam))
        
        self.__label = np.array(label) 
    
    def fit_predict(self,X:np.ndarray,y:np.ndarray,max_feature=None,min_fitur=0):
        self.fit(X,y,max_feature,min_fitur)
        return self.__label
    
    def predict(self,X:np.ndarray):
        if len(X.shape) ==1:
            return self.predict_tree(self.__tree,X)
        else:
            label = []
            for sam in X:
                label.append(self.predict_tree(self.__tree,sam))
            return np.array(label)


    def predict_tree(self,node, sample):
    # Memprediksi label untuk sampel menggunakan pohon keputusan
        if node.result is not None:
            return node.result

        if sample[node.feature] <= node.value:
            return self.predict_tree(node.left, sample)
        else:
            return self.predict_tree(node.right, sample)
    
    def score_accuracy(self,y_pred:np.ndarray,y_true:np.ndarray):
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true/len(y_pred)


class RegresiLinear:
    
    def __init__(self):
        self.__weight = 0
        self.__reg = None
    
    @property
    def reg(self):
        #hanya dekorator
        pass

    @reg.getter
    def reg_(self):
        return self.__reg
    
    def fit(self,x:np.ndarray,y:np.ndarray):
        x_T = np.transpose(x)
        multiply = x_T @ x 
        try:
            inverse = np.linalg.inv(multiply)
            self.__weight = inverse @ x_T @ y
        except np.linalg.LinAlgError:
            print("The matrix is singular and not invertible.")
        
    def fit_predict(self, x:np.ndarray, y:np.ndarray):
        self.fit(x,y)
        self.__reg = x @ self.__weight
        return self.__reg
    
    def predict(self, x:np.ndarray):
        return x @ self.__weight



    



from setuptools import find_packages, setup

with open("INDOML/Readme.md","r") as f:
    deskripsi = f.read()

setup(
    name="INDOML",
    version = "0.0.30",
    description="package machine learning buatan indonesia",
    package_dir={"":"INDOML"},
    packages=find_packages(where='INDOML'),
    long_description=deskripsi,
    long_description_content_type='text/markdown',
    url = 'https://github.com/khalifardy/library_machine_learning',
    author = 'Miko',
    author_email= 'khalifardy.miqdarsah@gmail.com',
    license= 'GNU',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
    ],
    install_requires = [
        'numpy >= 1.26.2',
        'seaborn >= 0.13.0',
        'matplotlib >= 3.8.2',
        'scikit-learn >= 1.3.2',
        'pandas >= 2.1.4'
    ],
    python_requires = ">=3.10"
)

